package com.akman.skintone;
import static jcuda.driver.JCudaDriver.cuCtxCreate;
import static jcuda.driver.JCudaDriver.cuCtxSynchronize;
import static jcuda.driver.JCudaDriver.cuDeviceGet;
import static jcuda.driver.JCudaDriver.cuInit;
import static jcuda.driver.JCudaDriver.cuLaunchKernel;
import static jcuda.driver.JCudaDriver.cuMemAlloc;
import static jcuda.driver.JCudaDriver.cuMemFree;
import static jcuda.driver.JCudaDriver.cuMemcpyDtoH;
import static jcuda.driver.JCudaDriver.cuMemcpyHtoD;

import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;

import ij.IJ;
import jcuda.Pointer;
import jcuda.Sizeof;
import jcuda.driver.CUcontext;
import jcuda.driver.CUdevice;
import jcuda.driver.CUdeviceptr;
import jcuda.driver.CUfunction;
import jcuda.driver.JCudaDriver;
import jcuda.nvrtc.JNvrtc;

public class SkinTone {

	public static void calculateSkintoneGPU() throws IOException {
		BufferedImage bufferedImage = ImageIO.read(SkinTone.class.getClassLoader().getResource("lena512color.png"));
		if (bufferedImage == null || bufferedImage.getData() == null)
			return;
		Raster raster = bufferedImage.getData();
		
		DataBuffer dataBuffer = raster.getDataBuffer();
        DataBufferByte dataBufferInt = (DataBufferByte)dataBuffer;
        byte[] pixels =  dataBufferInt.getData();
        
		int totalPixels = raster.getHeight() * raster.getWidth();

		CUfunction kernelFunction = initlize();

		int output[] = execute(kernelFunction, pixels, raster.getWidth(), raster.getHeight());
		// Flushing memory
		raster = null;
		bufferedImage.flush();
		bufferedImage = null;

		long skintoneThreshold = Math.round(output[0] / (double) totalPixels * 100.0);

		System.err.println("Skintone Using GPU: " + output[0]);
		System.err.println("Total Pixel Of GPU: " + totalPixels);
		System.err.println("SKinTone Percentage Using GPU: " + skintoneThreshold + "%");
	}

	static int[] execute(CUfunction kernelFunction, byte[] pixels, int w, int h) {
		// Allocate memory on the device, and copy the host data to the device
		int size = w * h * Sizeof.INT;
		CUdeviceptr pointer = new CUdeviceptr();
		cuMemAlloc(pointer, size);
		cuMemcpyHtoD(pointer, Pointer.to(pixels), size);

		int numElements = 9;

		// Allocate device output memory
		CUdeviceptr deviceOutput = new CUdeviceptr();
		cuMemAlloc(deviceOutput, numElements * Sizeof.INT);

		// Set up the kernel parameters: A pointer to an array
		// of pointers which point to the actual values.
		Pointer kernelParameters = Pointer.to(Pointer.to(pointer), Pointer.to(new int[] { w }),
				Pointer.to(new int[] { h }), Pointer.to(deviceOutput));

		// Call the kernel function
		int blockSize = 16;
		int gridSize = (Math.max(w, h) + blockSize - 1) / blockSize;
		cuLaunchKernel(kernelFunction, gridSize, gridSize, 1, // Grid dimension
				blockSize, blockSize, 1, // Block dimension
				0, null, // Shared memory size and stream
				kernelParameters, null // Kernel- and extra parameters
		);
		cuCtxSynchronize();

		// Allocate host output memory and copy the device output
		// to the host.
		int hostOutput[] = new int[numElements];
		cuMemcpyDtoH(Pointer.to(hostOutput), deviceOutput, numElements * Sizeof.INT);

		// Clean up.
		cuMemFree(deviceOutput);
		cuMemFree(pointer);

		return hostOutput;
	}

	public static CUfunction initlize() {
		System.out.println(" Loading Cuda Device and Code");

		// Enable exceptions and omit all subsequent error checks
		JCudaDriver.setExceptionsEnabled(true);
		JNvrtc.setExceptionsEnabled(true);

		// Initialize the driver and create a context for the first device.
		cuInit(0);
		CUdevice device = new CUdevice();
		cuDeviceGet(device, 0);
		CUcontext context = new CUcontext();
		cuCtxCreate(context, 0, device);

		// Obtain the CUDA source code from the CUDA file
		String cuFileName = "Skintone.cu";
		String sourceCode = CudaUtils.readResourceAsString(cuFileName);
		if (sourceCode == null) {
			IJ.showMessage("Error", "Could not read the kernel source code");
		}

		// Create the kernel function
		return CudaUtils.createFunction(sourceCode, "skintone");
	}

	public static void calculateSkintoneCPU() throws IOException {
		BufferedImage bufferedImage = ImageIO.read(SkinTone.class.getClassLoader().getResource("lena512color.png"));
		if (bufferedImage == null || bufferedImage.getData() == null)
			return;
		Raster raster = bufferedImage.getData();
		float[] rgb = new float[4];
		int totalPixels = raster.getHeight() * raster.getWidth();

		int skinTonePixels = 0;

		for (int x = 0; x < raster.getWidth(); x++) {
			for (int y = 0; y < raster.getHeight(); y++) {
				raster.getPixel(x, y, rgb);
				if (skintone(rgb)) {
					skinTonePixels++;
				}
			}
		}

		// Flushing memory
		raster = null;
		rgb = null;
		bufferedImage.flush();
		bufferedImage = null;

		long skintoneThreshold = Math.round(skinTonePixels / (double) totalPixels * 100.0);

		System.err.println("Skintone Using CPU: " + skinTonePixels);
		System.err.println("Total Pixel Of CPU: " + totalPixels);
		System.err.println("SKinTone Percentage Using CPU: " + skintoneThreshold + "%");
	}

	private static boolean skintone(float[] rgb) {
		float yCbCr[] = (float[]) convertRGBtoYUV(rgb);
		if ((yCbCr[1] >= 80 && yCbCr[1] <= 120) && (yCbCr[2] >= 133 && yCbCr[2] <= 173)) {
			return true;
		}
		return false;
	}

	private static float[] convertRGBtoYUV(float[] rgb) {
		final float[] yCbCr = new float[3];
		float r = rgb[0];
		float g = rgb[1];
		float b = rgb[2];

		yCbCr[0] = 16 + (0.299f * r) + (0.587f * g) + (0.144f * b);
		yCbCr[1] = 128 + (-0.169f * r) - (0.331f * g) + (0.5f * b);
		yCbCr[2] = 128 + (0.5f * r) - (0.419f * g) - (0.081f * b);

		return yCbCr;
	}

}
